import sys


# Begin -- grammar generated by Yapps
import sys, re
from yapps import runtime

class HalScanner(runtime.Scanner):
    patterns = [
        ('"no"', re.compile('no')),
        ('"yes"', re.compile('yes')),
        ('"option"', re.compile('option')),
        ('"function"', re.compile('function')),
        ('"param"', re.compile('param')),
        ('"pin"', re.compile('pin')),
        ('";"', re.compile(';')),
        ('"component"', re.compile('component')),
        ('"$"', re.compile('$')),
        ('//.*', re.compile('//.*')),
        ('/[*](.|\n)*?[*]/', re.compile('/[*](.|\n)*?[*]/')),
        ('[ \t\r\n]+', re.compile('[ \t\r\n]+')),
        ('END', re.compile(';;')),
        ('DIRECTION', re.compile('r|w|rw')),
        ('TYPE', re.compile('float|bit|u32|s32|u16|s16|u8|s8')),
        ('FP', re.compile('fp|nofp')),
        ('NAME', re.compile('[a-zA-Z_][a-zA-Z0-9_]*')),
    ]
    def __init__(self, str,*args,**kw):
        runtime.Scanner.__init__(self,None,{'[ \t\r\n]+':None,'//.*':None,'/[*](.|\n)*?[*]/':None,},str,*args,**kw)

class Hal(runtime.Parser):
    Context = runtime.Context
    def File(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'File', [])
        while self._peek('"$"', '"component"', '"pin"', '"param"', '"function"', '"option"', context=_context) != '"$"':
            Declaration = self.Declaration(_context)
        self._scan('"$"', context=_context)
        return True

    def Declaration(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'Declaration', [])
        _token = self._peek('"component"', '"pin"', '"param"', '"function"', '"option"', context=_context)
        if _token == '"component"':
            self._scan('"component"', context=_context)
            NAME = self._scan('NAME', context=_context)
            self._scan('";"', context=_context)
            comp(NAME);
        elif _token == '"pin"':
            self._scan('"pin"', context=_context)
            NAME = self._scan('NAME', context=_context)
            TYPE = self._scan('TYPE', context=_context)
            DIRECTION = self._scan('DIRECTION', context=_context)
            self._scan('";"', context=_context)
            pin(NAME, TYPE, DIRECTION)
        elif _token == '"param"':
            self._scan('"param"', context=_context)
            NAME = self._scan('NAME', context=_context)
            TYPE = self._scan('TYPE', context=_context)
            DIRECTION = self._scan('DIRECTION', context=_context)
            self._scan('";"', context=_context)
            param(NAME, TYPE, DIRECTION)
        elif _token == '"function"':
            self._scan('"function"', context=_context)
            NAME = self._scan('NAME', context=_context)
            if self._peek('FP', '";"', context=_context) == 'FP':
                FP = self._scan('FP', context=_context)
            self._scan('";"', context=_context)
            function(NAME, FP)
        else: # == '"option"'
            self._scan('"option"', context=_context)
            NAME = self._scan('NAME', context=_context)
            Value = self.Value(_context)
            self._scan('";"', context=_context)
            option(NAME, Value)

    def Value(self, _parent=None):
        _context = self.Context(_parent, self._scanner, 'Value', [])
        _token = self._peek('"yes"', '"no"', context=_context)
        if _token == '"yes"':
            self._scan('"yes"', context=_context)
            return 1
        else: # == '"no"'
            self._scan('"no"', context=_context)
            return 0


def parse(rule, text):
    P = Hal(HalScanner(text))
    return runtime.wrap_error_reporter(P, rule)

# End -- grammar generated by Yapps



dirmap = {'r': 'HAL_RD', 'w': 'HAL_WR', 'rw': 'HAL_RD_WR' }

functions = []; params = []; pins = []; options = {}
comp_name = None

def comp(name):
    global comp_name
    comp_name = name;

def pin(name, type, dir):
    pins.append((name, type, dir))
def param(name, type, dir):
    params.append((name, type, dir))
def function(name, fp):
    functions.append((name, fp == "fp"))
def option(name, value):
    options[name] = value

def prologue(f):
    print >> f, """\
#include "rtapi.h"
#include "rtapi_app.h"
#include "hal.h"

static int comp_id;

#define FUNCTION(name) static void name(struct state *arg, long period)
"""
    names = {}

    print >>f
    print >>f, "struct state {"
    for name, type, dir in pins:
        if names.has_key(name):
            raise RuntimeError, "Duplicate pin name: %s" % name
        print >>f, "    hal_%s_t *%s;" % (type, name)
        names[name] = 1

    for name, type, dir in params:
        if names.has_key(name):
            raise RuntimeError, "Duplicate param name: %s" % name
        print >>f, "    hal_%s_t %s;" % (type, name)
        names[name] = 1

    print >>f, "};"
    
    print >>f
    names = {}
    for name, fp in functions:
        if names.has_key(name):
            raise RuntimeError, "Duplicate function name: %s" % name
        print >>f, "static void %s(struct state *arg, long period);" % name
        names[name] = 1


    print >>f
    print >>f, "static int export(char *prefix, struct state *arg) {"
    print >>f, "    char buf[HAL_NAME_LEN + 2];"
    print >>f, "    int r = 0;"
    for name, type, dir in pins:
        print >>f, "    r = hal_pin_%s_newf(%s, &(arg->%s), comp_id," % (
            type, dirmap[dir], name)
        print >>f, "        \"%%s.%s\", prefix);" % (name.replace("_", "-"))
        print >>f, "    if(r != 0) return r;"

    for name, type, dir in params:
        print >>f, "    r = hal_param_%s_newf(%s, &(arg->%s), comp_id," % (
            type, dirmap[dir], name)
        print >>f, "        \"%%s.%s\", prefix);" % (name.replace("_", "-"))
        print >>f, "    if(r != 0) return r;"

    for name, fp in functions:
        print >>f, "    rtapi_snprintf(buf, HAL_NAME_LEN, \"%%s.%s\", prefix);"\
            % name
        print >>f, "    r = hal_export_funct(buf, (void(*)(void *arg, long))%s, arg, %s, 0, comp_id);" % (
            name, int(fp))
        print >>f, "    if(r != 0) return r;"
    print >>f, "    return 0;"
    print >>f, "}"

    if options.get("rtapi_app", 1):
        if not options.get("singleton"):
            print >>f, "static int count = 1;"
            print >>f, "MODULE_PARM(count, \"i\");"
            print >>f, "MODULE_PARM_DESC(count, \"number of %s\");" % comp_name

        print >>f, "int rtapi_app_main(void) {"
        print >>f, "    int r = 0;"
        print >>f, "    struct state *states = NULL;"
        if not options.get("singleton"):
            print >>f, "    int i;"

        print >>f, "    comp_id = hal_init(\"%s\");" % comp_name
        print >>f, "    if(comp_id < 0) return comp_id;"

        if options.get("singleton"):
            print >>f, "    states = hal_malloc(sizeof(struct state));"
            print >>f, "    r = export(\"%s\", states);" % \
                    comp_name.replace("_", "-")
        else:
            print >>f, "    states = hal_malloc(count*sizeof(struct state));"
            print >>f, "    for(i=0; i<count; i++) {"
            print >>f, "        char buf[HAL_NAME_LEN + 2];"
            print >>f, "        rtapi_snprintf(buf, HAL_NAME_LEN, " \
                                        "\"%s.%%d\", i);" % \
                                    comp_name.replace("_", "-")
            print >>f, "        r = export(buf, &states[i]);"
            print >>f, "        if(r != 0) break;"
            print >>f, "    }"
        print >>f, "    if(r) hal_exit(comp_id);"
        print >>f, "    return r;";
        print >>f, "}"

        print >>f
        print >>f, "void rtapi_app_exit(void) { hal_exit(comp_id); }"

    if not options.get("no_convenience_defines"):
        print >>f
        for name, type, dir in pins:
            print >>f, "#define %s (*arg->%s)" % (name, name)
        for name, type, dir in params:
            print >>f, "#define %s (arg->%s)" % (name, name)
        
    print >>f
    print >>f

import sys
if len(sys.argv) > 1:
    f = open(sys.argv[1])
else:
    f = sys.stdin

f = f.read()
a, b = f.split("\n;;\n", 1)

p = parse('File', a)
if not p: raise SyntaxError

if len(sys.argv) > 2:
    f = open(sys.argv[2], "w")
else:
    f = sys.stdout

prologue(f)
f.write(b)
